package ca;

import ca.models.*;
import java.awt.*;

public class RoadCanvas extends Canvas implements Runnable {   
  public Road freeway;
  private Image buffer;
  private long TIMESTEP = 100;
  private Graphics bg = null;
  private Thread runner=null;
  private Dimension minSize = new Dimension (600, 500);
  // Dinstance between two updates of the diagrams
  private final int UPDATE_DIAGRAM = 10;

  // diagram labels  
  private final String SpaceLabel = "space";
  private final String[] TimeLabel = {"t","i","m","e"};

  private final String x1="rho";
  private final String y1="q";
  private final String x2="rho";
  private final String y2="v";
  private final String x3="q";
  private final String y3="v";
  private final String x4="v";
  private final String y4="f";
  private final String x5="gap";
  private final String y5="f";

  private final String[] xAxisLabel={x1,x2,x3,x4,x5};
  private final String[] yAxisLabel={y1,y2,y3,y4,y5};
  
  // dot sizes
  private final int DOTSIZE = 1;   
  private final int XDOTDIST = 1;

  // counter of the rows in the space-time-plot
  private int row;

  // some properties of the several graphic outputs
  private double density;
  private double slowdown;
  private double lambda;
  private double deltaX;

  private int xsize;
  private int ysize;

  private int xsizeSTD;
  private int ysizeSTD;
  private int xsizeIndy;
  private int ysizeIndy;
  private int xsizeDiagram;
  private int ysizeDiagram;
  private int xsizeDiagramPart;
  private int ysizeDiagramPart;
  
  private final int xShift = 17;
  private final int yShift = 17;
  private final int bluebar = 3;
   
  // some properties of the indianapolis scenario
  private int radIn;
  private int radOut;
  private int radVeh;
  private int xIndyMid;
  private int yIndyMid;
  
  // some properties of the diagrams
  private int[] xDiagram;
  private int[] yDiagram;

  private int[] xOrigin;
  private int[] yOrigin;
  
  private double[] xDelta;
  private double[] yDelta;

  // my private colors
  private final Color streetcolor = Color.lightGray;
  private final Color background = Color.white;
  private final Color foreground = Color.black;
  private final Color linecolor = Color.blue;
  private final Color vcolor[];

  RoadCanvas(double density, double slowdown, double lambda, 
	     double deltaX, Simulator sim) {

    this.density = density;
    this.slowdown = slowdown;
    this.lambda = lambda;
    this.deltaX = deltaX;
 
    freeway = new Road (density, deltaX, sim);
    vcolor = new Color[freeway.vfreq.length];

    row = 0;

    // colors for different velocities
    for (int i= 0; i<vcolor.length; i++) {
      switch(i) {
      case 0: vcolor[i] = Color.red;             break;
      case 1: vcolor[i] = new Color(205,69,0);   break;
      case 2: vcolor[i] = Color.orange;          break;
      case 3: vcolor[i] = new Color(154,205,50); break;
      case 4: vcolor[i] = new Color(84,255,159); break;
      case 5: vcolor[i] = new Color(0,255,127);  break;
      default: 
	int grey = (200*i)/freeway.vfreq.length;
	vcolor[i] = new Color (grey, grey, grey);
      }
    }
  }

  public void newValues (double density, double slowdown, double lambda) {
    this.density = density;
    this.slowdown = slowdown;
    this.lambda = lambda;
  }

  public void run() {  
    long startTime, wait=0;

    // System.out.println ("Thread running");

    startTime = System.currentTimeMillis();

    // loop counter
    // shows, whether the diagrams have to be plotted
    int counter = 0;

    // Just to be nice, lower this thread's priority
    // so it can't interfere with other processing going on.
    // Thread.currentThread().setPriority(Thread.MIN_PRIORITY);

    while ((Thread.currentThread() == runner) &&
           (++counter < 2*UPDATE_DIAGRAM)) {
      if (bg == null) { 
	xsize = size().width;
	ysize = size().height;
	
	// System.out.println ("new buffer (" + xsize + "," + ysize + ")");
	
	xsizeSTD = xsize/2;
	ysizeSTD = ysize/2;
	xsizeIndy = xsize/2;
	ysizeIndy = ysize/2;
	xsizeDiagram = xsize/2;
	ysizeDiagram = ysize;
	xsizeDiagramPart = xsizeDiagram/8;
	ysizeDiagramPart = ysize/40;
	
	radIn = (int)(0.34*xsizeIndy);
	radOut = (int)(0.38*xsizeIndy);
	radVeh = (int)(0.36*xsizeIndy);
	
	xIndyMid = xsizeIndy/2+xShift+bluebar;
	yIndyMid = ysizeSTD+ysizeIndy/2+yShift/2;
	
	xDiagram = new int[5];
	yDiagram = new int[5];
	xOrigin = new int[5];
	yOrigin = new int[5];
	xDelta = new double[5];
	yDelta = new double[5];
	
	for (int i=0;i<3;i++) {
	  xDiagram[i] = xsizeDiagram-2*xsizeDiagramPart;
	  yDiagram[i] = (ysizeDiagram-4*ysizeDiagramPart)/4;
	  xOrigin[i] = xsizeSTD+xsizeDiagramPart;
	  yOrigin[i] = (i+1)*(yDiagram[i]+ysizeDiagramPart);
	  xDelta[i] = 1.0;
	  yDelta[i] = 1.0;
	}
	
	for (int i=3;i<5;i++) {
	  xDiagram[i] = (int)(0.44*xDiagram[0]);
	  yDiagram[i] = (ysizeDiagram-4*ysizeDiagramPart)/4;
	  xOrigin[i] = (i == 3) ? xsizeSTD+xsizeDiagramPart : 
	    xsizeSTD+xsizeDiagram/2;
	  yOrigin[i] = 4*(yDiagram[i]+ysizeDiagramPart)-8;
	  xDelta[i] = 1.0;
	  yDelta[i] = 1.0;
	}
	
	// draw all diagrams
	buffer = createImage(xsize,ysize);
	bg = buffer.getGraphics();
	bg.setColor (getBackground());
        bg.fillRect (0, 0, xsize, ysize);
	bg.clearRect(xShift+bluebar,DOTSIZE+yShift,xsizeSTD,ysizeSTD);
	bg.setColor (getForeground());
	
	Font oldFont = bg.getFont();

	bg.setFont (new Font (oldFont.getName(), Font.ITALIC, 
			      oldFont.getSize())); 

	for (int i=0;i<5;i++) {
	  diagramAxis(bg,xOrigin[i],yOrigin[i],
		      xOrigin[i]+xDiagram[i],yOrigin[i]);
	  bg.drawString(xAxisLabel[i],xOrigin[i]+xDiagram[i]+3,yOrigin[i]+5);
	  
	  diagramAxis(bg,xOrigin[i],yOrigin[i],
		      xOrigin[i],yOrigin[i]-yDiagram[i]);
	  bg.drawString(yAxisLabel[i],xOrigin[i]-10,yOrigin[i]-yDiagram[i]+25);
	}
	bg.setFont (oldFont);

	drawstreet(bg,xIndyMid,yIndyMid,radOut,radIn);
	bluelines(bg,xShift,yShift,bluebar-1,ysizeSTD);
	SpaceTimeLabels(bg,xsizeSTD/3,3*yShift/4,xsizeSTD/2,yShift/2,50,
			SpaceLabel,xShift/2,yShift+ysizeSTD/4,
			xShift/2,yShift+ysizeSTD/2,50,TimeLabel);
	ClearDiagrams();
      }
      
      freeway.update (slowdown, density, lambda);

      // update space-time-plot
      spacetimepaint(bg,xShift+row,XDOTDIST,DOTSIZE,xShift+bluebar);
      // update indy-scenario
      indypaint(bg,DOTSIZE*2,xIndyMid,yIndyMid,radVeh);
      
      // if enabled, plot diagrams
      if (counter%2 == 0) {	
	freeway.measure();
	for (int i=0;i<((counter == 6)? 5 : 3);i++) {
	  diagram(bg,xOrigin[i],yOrigin[i],xDiagram[i],yDiagram[i],
		  xDelta[i],yDelta[i],i);
	}
      }
    
      // move space-time-plot upwards
      if (row < ysizeSTD-DOTSIZE) {
	row+=DOTSIZE;
      } else { 
	bg.copyArea(xShift+bluebar,DOTSIZE+yShift,xsizeSTD,ysizeSTD-DOTSIZE,
		    0,-DOTSIZE);
	bg.clearRect(xShift+bluebar,ysizeSTD-DOTSIZE+yShift,xsizeSTD,DOTSIZE);
      }

      if (counter >= UPDATE_DIAGRAM)
	counter = 0;
      
      repaint();
      
      startTime += TIMESTEP;
      wait = startTime-System.currentTimeMillis();
      if (wait > 0) {   
	try {Thread.sleep (wait);} catch(InterruptedException e) {}
      }
    }
    // System.out.println ("Thread dying");
  }

  //////////////////
  // space-time-plot
  
  private void spacetimepaint(Graphics g, int row,
			      int dotdist, int dotsize, int xshift) {  
    // we only have space for 300 points. So if we have more cells than
    // that we need to combine multiple cells into one
    int combine = freeway.LENGTH/300; // number of cells to be combined
    int i, j, num, speed;

    for (i = 0; i < 300; i++) {
      num = 0; speed = 0;
      for (j = 0; j < combine; j++) {
	if (freeway.speed[i*combine+j] >= 0) {
	  speed += freeway.speed[i*combine+j];
	  num++;
	}
      }
      if (num > 0) {
	g.setColor(vcolor[speed/num]);   
	g.fillRect((xshift+i)*dotdist,row,dotsize,dotsize);
      }
    }
    g.setColor(background);   
  }

  /////////////////////////////////
  // plot the indianapolis-scenario
   
  private void indypaint(Graphics g, int dotsize,
			 int xm, int ym, int radius) {  
    // clear road
    drawstreet (g, xIndyMid, yIndyMid, radOut, radIn);

    // degrees per vehicle
    double f = 2.0*Math.PI/(double)freeway.LENGTH;
    for (int i = 0; i < freeway.LENGTH; i++) {
      // angle of a vehicle
      double rad = f*i;
      if (freeway.speed[i] != -1) {
	// plot a vehicle
	g.setColor(foreground);
	g.fillRect((int)(xm+Math.cos(rad)*radius), 
		   (int)(ym+Math.sin(rad)*radius), 
		   dotsize, dotsize);
      } 
      // else {
      // Erase a vehicle
      // g.setColor(streetcolor);
      // g.fillRect((int)(xm+Math.cos(rad)*radius), 
      //            (int)(ym+Math.sin(rad)*radius), 
      //  	   dotsize, dotsize);
      // }
    }
  }

  ////////////////
  // plot diagrams

  void diagram(Graphics g, int x1, int y1, int intdx, int intdy, 
	       double dbldx, double dbldy, int index) {
    double xValue;
    double yValue;
    g.setColor(foreground);
    // fundamental diagrams
    if (index < 3) {
      switch(index) {
      case 0: 
	// local flow = local flow(local density)
	xValue = freeway.dens*7.5/deltaX;
	yValue = freeway.flow;
	break;
      case 1:
	// local mean velocity = local mean velocity(local density)
	xValue = freeway.dens*7.5/deltaX;
	yValue = freeway.v/freeway.MAXSPEED;
	break;
      default:
	// local mean velocity = local mean velocity(local flow)
	xValue = freeway.flow;
	yValue = freeway.v/freeway.MAXSPEED;
	break;
      }
      if ((xValue >= 0) && (xValue/dbldx <= 1.0) &&
	  (yValue >= 0) && (yValue/dbldy <= 1.0)) {
	// System.out.println ("x: " + xValue + " dbldx: " + dbldx + 
	//		    " intdx: " + intdx);
	g.drawRect(x1+(int)(xValue/dbldx*intdx),
		   y1-(int)(yValue/dbldy*intdy),1,1);
      }
    } else {
      // bar charts

      // bar chart of the frequency of velocities
      if (index == 3) {
	// clear diagram
	g.setColor(background);
	g.fillRect(x1+1,y1-intdy,intdx-1,intdy);
	drawArrow(g,x1,y1-intdy,10,8,270,true,foreground);
	drawArrow(g,x1+intdx,y1,10,8,0,true,foreground);
	// bar widht
	int bar = intdx/(freeway.MAXSPEED+1);
	g.setColor(foreground);
	for (int i=0;i<=freeway.MAXSPEED;i++) {
	  g.setColor(vcolor[i]);
	  int y = (int)(freeway.vfreq[i]/dbldy*intdy);
	  g.fillRect(x1+i*bar+1,y1-y,bar-1,y);
	}
      }	else {
	// bar chart of the frequency of gaps
	
	// clear diagram
	g.setColor(background);
	g.fillRect(x1+1,y1-intdy,intdx-1,intdy);
	drawArrow(g,x1,y1-intdy,10,8,270,true,foreground);
	drawArrow(g,x1+intdx,y1,10,8,0,true,foreground); 
	// number of bars
	int xmax = 18;
	// bar width
	int bar = intdx/xmax;
	g.setColor(foreground);
	for (int i=0;i<xmax;i++) {
	  int y = (int)(freeway.gapfreq[i]/dbldy*intdy);
	  g.fillRect(x1+i*bar+1,y1-y,bar-1,y);
	}
      }
    }
  }
  
  ///////////////
  // diagram axes
  
  private void diagramAxis(Graphics g, int xOr, int yOr, 
			   int xDiag, int yDiag) {
    g.drawLine(xOr,yOr,xDiag,yDiag);
    drawArrow(g,xDiag,yDiag,10,8,(xOr == xDiag)? 270: 0,true,foreground);
  }

  /////////////////////////////////////////////
  // draw the road of the indianapolis-scenario
  
  private void drawstreet(Graphics g, int xm, int ym, int r1, int r2) {
    g.setColor(streetcolor);
    g.fillOval(xm-r1,ym-r1,2*r1,2*r1);
    g.setColor(background);
    g.fillOval(xm-r2,ym-r2,2*r2,2*r2);
    g.setColor(linecolor);
    g.drawLine(xm+r2-20,ym,xm+r1+20,ym);
    g.setColor(foreground);
    g.drawArc(xm-r2+30,ym-r2+30,2*r2-60,2*r2-60,0,-270);
    drawArrow(g, xm,ym-r2+30,10,8,355, true, Color.black);
  }


  //////////////////////////////
  // blue line at the first site

  private void bluelines(Graphics g,int x1,int y1,int x2,int y2) {  
    Color old_color = g.getColor();
    g.setColor(linecolor);
    g.fillRect(x1,y1,x2,y2);
    g.setColor(old_color);
  }
  
  ///////////////////////////////////////////////////////
  // Labels of the space-time-plot
  // The letters of the ordinate are arranged vertically.

  private void SpaceTimeLabels(Graphics g,int x1,int y1,int xA1,int yA1,
			       int length1,String label1,int x2,int y2,
			      int xA2,int yA2,int length2,String[] label2) {  
    Color color = foreground;
    Color old_color = g.getColor();
    // Font oldFont = g.getFont();

    // g.setFont (new Font (oldFont.getName(), Font.ITALIC, 
    //                      oldFont.getSize())); 
    g.setColor(color);
    g.drawString(label1,x1,y1);
    g.drawLine(xA1,yA1,xA1+length1,yA1);
    drawArrow(g,xA1+length1,yA1,10,8,0,true,color);
    for (int i=0;i<4;i++) {
      g.drawString(label2[i],x2-5,y2+i*15);
    }
    g.drawLine(xA2,yA2,xA2,yA2+length2);
    drawArrow(g,xA2,yA2+length2,10,8,90,true,color);
    g.setColor(old_color);
    // g.setFont (oldFont);
  }

  //////////////////////////////////////////////////////////
  // plot an arrow with any properties like size, angle, ...

  private void drawArrow(Graphics g,int xA,int yA,int Length,int Width,
			int deg,boolean filled,Color color) {
    
    double Angle=(double)deg/180.0*Math.PI;
    double sinA = Math.sin(Angle);
    double cosA = Math.cos(Angle);
    int x1 = xA-(int)(cosA*Length + sinA*Width/2);
    int y1 = yA-(int)(sinA*Length - cosA*Width/2);
    
    Polygon filledPolygon = new Polygon();
    filledPolygon.addPoint(xA, yA);
    filledPolygon.addPoint(x1,y1);
    filledPolygon.addPoint(x1+(int)(sinA*Width),y1-(int)(cosA*Width));
    Color old_color = g.getColor();
    g.setColor(color);
    if(filled) {
      g.fillPolygon(filledPolygon); 
    } else {
      g.drawPolygon(filledPolygon); 
    }
    g.setColor(old_color);
    
  }

  /////////////////
  // clear diagrams

  public void ClearDiagrams() {
    bg.setColor(background);
    for (int i=0;i<5;i++) {
      bg.fillRect(xOrigin[i]+1,yOrigin[i]-yDiagram[i],
		  xDiagram[i]-1,yDiagram[i]);
      drawArrow(bg,xOrigin[i],yOrigin[i]-yDiagram[i],
		10,8,270,true,foreground);
      drawArrow(bg,xOrigin[i]+xDiagram[i],yOrigin[i],
		10,8,0,true,foreground);
    }
    bg.setColor(foreground);
    repaint();
  }

  public void start() {  
    // System.out.println ("start()");
    if (runner == null) {
      // System.out.println ("starting Thread");

      runner = new Thread (this);
      runner.start();
    }
    // System.out.println ("done with start()");
  }
  
  public void stop() { 
    // System.out.println ("stop()");
    if (runner != null) {
      // System.out.println ("stopping Thread");

      Thread waitFor = runner;
      // Stop the animating thread.
      runner = null;
      // wait for it to terminate
      // the timeout is for a bug in most versions of Netscape which hangs
      // after some time I can be reasonably sure that the thread has died
      try {waitFor.join(1000+2*TIMESTEP);} catch (InterruptedException e) {}
    }
    // System.out.println ("done with stop()");
  }

  public void paint (Graphics g) {  
    if (buffer != null) {
      // System.out.println ("draw buffer");
      g.drawImage (buffer, 0, 0, null);  
    }
  }
  
  public void update(Graphics g) {  
    paint(g);
  }

  public Dimension preferredSize() {
    return minimumSize();
  }

  public Dimension minimumSize() {
    return minSize;
  }

}  // End of "class RoadCanvas extends Canvas"
